package radioai

import (
	"fmt"
	"io"
	"math/rand"
	"net/http"
	"net/http/httputil"
	"strconv"
	"strings"
	"sync"
	"time"
)

// Cache the list of active nodes
type activeNodesCache struct {
	client   *RadioAPI
	nodes    []string
	deadline time.Time
	lock     sync.Mutex
}

var activeNodesTtl = 500 * time.Millisecond

func newActiveNodesCache(client *RadioAPI) *activeNodesCache {
	return &activeNodesCache{
		client: client,
		nodes: []string{},
	}
}

func (anc *activeNodesCache) GetNodes() []string {
	anc.lock.Lock()
	defer anc.lock.Unlock()

	now := time.Now()
	if now.After(anc.deadline) {
		if nodes, err := anc.client.GetNodes(); err == nil {
			anc.nodes = nodes
			anc.deadline = now.Add(activeNodesTtl)
		}
	}
	return anc.nodes
}

// HTTP redirector.
//
// All user-facing traffic reaches the redirector first (this is
// where the first-level, high-ttl redirection points to).
//
// The purpose of the HTTP redirector is two-fold: sources will be
// proxied to the master icecast server, while clients will be served
// a .m3u file directly pointing at the relays.
//
type HttpRedirector struct {
	client    *RadioAPI
	nodeCache *activeNodesCache
}

func NewHttpRedirector(client *RadioAPI) *HttpRedirector {
	return &HttpRedirector{
		client: client,
		nodeCache: newActiveNodesCache(client),
	}
}

// Return an active node, chosen randomly.
func (h *HttpRedirector) pickActiveNode() string {
	nodes := h.nodeCache.GetNodes()
	if nodes != nil && len(nodes) > 0 {
		return nodes[rand.Intn(len(nodes))]
	}
	return ""
}

// Parse the request and extract the mount path.
func (h *HttpRedirector) getMount(r *http.Request) (*Mount, error) {
	path := r.URL.Path
	if strings.HasSuffix(path, ".m3u") {
		path = path[:len(path)-4]
	}
	return h.client.GetMount(path)
}

// Serve a response for a client connection to a relay.
func (h *HttpRedirector) serveRelay(w http.ResponseWriter, r *http.Request) {
	mount, err := h.getMount(r)
	if err != nil {
		http.Error(w, "Not Found", http.StatusNotFound)
		return
	}

	// Find an active node.
	relayAddr := h.pickActiveNode()
	if relayAddr == "" {
		http.Error(w, "No active nodes", http.StatusServiceUnavailable)
		return
	}

	// Create the m3u response.
	m3u := fmt.Sprintf("http://%s%s\n", relayAddr, mount.Name)
	w.Header().Set("Content-Length", strconv.Itoa(len(m3u)))
	w.Header().Set("Content-Type", "audio/x-mpegurl")
	io.WriteString(w, m3u)
}

func (h *HttpRedirector) serveSource(w http.ResponseWriter, r *http.Request) {
	mount, err := h.getMount(r)
	if err != nil {
		http.Error(w, "Not Found", http.StatusNotFound)
		return
	}

	// Find the current master node.
	masterAddr, err := h.client.GetMasterAddr()
	if err != nil {
		http.Error(w, err.Error(), http.StatusServiceUnavailable)
		return
	}

	// Proxy the resulting connection.
	proxy := &httputil.ReverseProxy{
		Director: func(req *http.Request) {
			req.URL.Host = masterAddr
			req.URL.Path = mount.Name
		},
		FlushInterval: time.Second,
	}
	proxy.ServeHTTP(w, r)
}

func (h *HttpRedirector) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	if r.Method == "SOURCE" {
		h.serveSource(w, r)
	} else {
		h.serveRelay(w, r)
	}
}
